#ifndef _SUT_DEFS
#define _SUT_DEFS

#include "patchlevel.h"

/* 1024 nodes */
#define SUT_MAX_LOGP 10

/* Size of buffer for moving data */
#define SUT_MAX_BUF 4096

/* System includes needed */
#include <stdio.h>
#include <signal.h>
#include <sys/time.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/param.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/errno.h>
#if !defined(SUT_cray)
extern int errno;
#endif

/* Node specification consists of Ranges, domains, and lists of nodes */
typedef struct { int start, end, stride; } SUT_Range;
typedef struct {
    int n;                 /* Number of nodes in domain */
    char *domain_name;     /* Name of domain (filename) */
    char **names;          /* Array of node names */
    } SUT_Domain;

typedef struct {
    int        nodenum;    /* Number relative to domain */
    SUT_Domain *domain;    /* Domain that node is in */
    char       *name;      /* Name of node */
    } SUT_Node_entry;

/* Messages are sent with a header.  Types of messages are

   op = SUT_STDxxx, len = bytes of data
   op = SUT_DONE, len = status   (process is ready to exit; status
   is status of user process)
   op = SUT_EXIT                 (parent is telling child to exit)
   op = SUT_ALLOW                (parent is telling child that it may
                                  write again)
   op = SUT_USER_SIG             (data contains number of signal)
   op = SUT_KILL                 (terminate all processes)
   op = SUT_NODELIST             (list of nodes for subtree)

   op = SUT_USER_ENV             (user's environment, not yet implemented)
 */
typedef enum { SUT_STDOUT = 1, SUT_STDIN = 2, SUT_STDERR = 3, 
	       SUT_DONE = 4, SUT_EXIT = 5, SUT_ALLOW = 6,
	       SUT_USER_SIG = 7, SUT_KILL = 8, 
	       SUT_ALIVE = 9, SUT_PROCEED = 10, SUT_NODELIST = 11
	       } SUT_Op;

typedef struct {
    SUT_Op op;
    int    len;
    } SUT_Header;

/* SUT_MAXOP is the maximum number of message-types (SUT_Op values from
   1 to 63, with 0 reserved for invalid) */
#define SUT_MAXOP 64
typedef struct {
    int (*(msghandler[SUT_MAXOP]))();
    } SUT_Opfcns;

typedef enum { SUT_UNOPEN, SUT_READY, SUT_STARTED, SUT_DEAD } SUT_Conn_state;
/* 
   This structure contains all of the information on handling a 
   connection.  Note that there are parent, child, and local process
   connections; all have the same structure (so that they can 
   all be handled by the same select operation)

   Special notes:
   The readfilter function is used to read data from the process.
   Currently, if not null, then the child does NOT know the stream
   message types, and must be handled separately.
 */
typedef struct {
    int readfd,                      /* fd for reading FROM process */
        writefd,                     /* fd for writing TO process */
        pid,                         /* pid of process */
        stat,                        /* stat value of process (on exit) */
        nodenum;                     /* Nodenum of process */

    /* Used to manage the actual connections and error recovery.
       Eventually, this will be used to allow the code to pick a
       new root from the list of (nchild,nlist). */
    SUT_Conn_state state;
    int nchild;                      /* Number of children */
    int *nlist;                      /* Host numbers of the children */

    /* Used to handle data read from process */
    char *linebuf;                   /* Buffer of line */
    int  linepos, linelen;           /* Current position and total length */
    int saw_newline;                 /* True if LAST character output
					 was a newline */
    
    SUT_Opfcns *ops;                 /* Pointer to structure with
					functions handlers (all siblings
					will point to same structure) */

    /* Each connection has a routine that is used to process input from the
       readfd.  If the source is a mangaer, this is SUT_Processinput.
       If it is a leaf, then it is a routine to add the host num and
       format as SUT_STDOUT.  If it is the user-root, it formats as
       SUT_STDIN.
     */
    int (*readdata)();               /* Routine to filter read data; may
					be null */
    int (*writedata)();
    /* Each connection has a routine to manage an EOF on the readfd.
       For leaves, this gets the exit status and sends it to the root;
       for managers, I've not decided exactly what it should do */
    int (*handle_eof)();
    } SUT_Connection;
    
typedef struct { 
    int            nactive;          /* Number of active connections */
    int            nconn;            /* Total number of connections */
    SUT_Connection conn[SUT_MAX_LOGP+2];
    int            parent;           /* index of connection of parent */
    int            local;            /* index of connection of local process */
    int            nchildren;        /* Number of children */
    int            children[SUT_MAX_LOGP];
                                     /* Connection number of children */
    fd_set         rdset;            /* fd_set for readfd from all 
					connections */
    int mynodenum;                   /* My node number (from the first
					range) */

    /* Old connection information; being phased out */
    int maxfds;                      /* Max fd (for select) */

    /* Info for output */
    int (*(writefilter))();          /* Routine to handle output */
    int (*(writeparent))();          /* Routine to write data to parent */
    int (*(readin))();               /* Routine to handle input */
    int availout;                    /* Amount of data that can be sent */
    int buflen;                      /* Length of each linebuf */

    /* Info for input */
    int (*(readparent))();           /* Routine to read from parent */
    /* Not yet handled ... */

    int retstat;                     /* Aggregate status return 
					for all child connections */
    } SUT_Jobs;

/* Remote shell program */
#ifndef SUT_REMSH
#define SUT_REMSH "rsh"
#endif

/* Timeout in seconds when starting remote program */
#ifndef SUT_TIMEOUT_VAL
#define SUT_TIMEOUT_VAL 20
#endif

/* Dynamic space allocation */
#define MALLOC(a) malloc((unsigned)(a))
#define FREE(a)   free((char *)a)
extern char *malloc();

/* Guarded system calls */
#define WRITE(a,str,n) {int _err; while ((_err = write(a,str,n)) != n) { \
    if (_err == -1 && errno == EINTR) continue; \
    fprintf( stderr, "Write failed in %s:%d\n", __FILE__, __LINE__ ); \
    if (_err < 0) perror( "Reason:" );}}
#define READ(a,str,len) {int _err; while ((_err = read(a,str,len)) != len) { \
    if (_err == -1 && errno == EINTR) continue; \
    fprintf( stderr, "Read failed in %s:%d\n", __FILE__, __LINE__ ); \
    if (_err < 0) perror( "Reason:" );}}

/* Debug bits */
extern int SUT_Debug;
extern int SUT_Mynode;
#define SUT_EXECVP     0x1
#define SUT_FORK       0x2
#define SUT_PRINT      0x4
#define SUT_HANDSHAKE  0x8
#define SUT_PASSDATA  0x10
#define SUT_EXITOPS   0x20
#ifdef SUT_DEBUG
#define DEBUG(a) a
#else
#define DEBUG(a)
#endif

/* Option bits */
extern int SUT_Options;
#define SUT_REGULATE   0x1

/* Known errors */
#define SUT_SUCCESS             0
#define SUT_RANGES_INVALID      1
#define SUT_NO_MEMORY           2
#define SUT_PIPE_FAILED         3
#define SUT_FORK_FAILED         4
#define SUT_SELECT_FAILED       5
#define SUT_GET_EXE_FAILED      6
#define SUT_EXEC_FAILED         7
#define SUT_HANDSHAKE_FAILED    8
#define SUT_INVALID_RANGE_SPEC  9
#define SUT_CHILD_UNRESPONSIVE 10
#define SUT_CHILD_CONFUSED     11
#define SUT_PARENT_CONFUSED    12
#define SUT_NOT_LOCAL          13    /* Not really an error, but needed */
#define SUT_CHILD_DONE         14    /* Not really an error, but needed */
#define SUT_UNEXPECTED_EOF     15
#define SUT_READ_FAILED        16

/* Bindings */
#ifndef ANSI_ARGS
#ifdef __STDC__
#define ANSI_ARGS(a) a
#else
#define ANSI_ARGS(a) ()
#endif
#endif
extern int SUT_Get_host_list ANSI_ARGS(( SUT_Range *, int, int **, int *));
extern int SUT_Get_exec_args ANSI_ARGS((int, char *(*)(), void *, 
					char **(*)(), void *,
					char *, char ***, int));
extern void SUT_Kill_child ANSI_ARGS(( int, int, char * ));
extern int SUT_Get_ranges ANSI_ARGS(( int, char **, SUT_Range **, int * ));
extern int SUT_Handshake ANSI_ARGS(( int, int, int ));
extern char *SUT_Condense ANSI_ARGS(( int *, int ));
extern int SUT_Getdebug ANSI_ARGS ((void));
extern int SUT_fork ANSI_ARGS ((void));
extern int SUT_execvp ANSI_ARGS ((char *, char **));
extern int SUT_CopyToOut ANSI_ARGS ((SUT_Jobs *, int));
extern int SUT_Write ANSI_ARGS ((SUT_Jobs *, int, int ));
extern int SUT_Writedata ANSI_ARGS (( int, char *, int ));
extern int SUT_Writedataroot ANSI_ARGS (( int, char *, int ));
extern SUT_Opfcns *SUT_CreateOpfcns ANSI_ARGS ((void));
extern int SUT_ProcessInput ANSI_ARGS (( SUT_Jobs *, int ));
extern int SUT_Leaf_exit ANSI_ARGS (( SUT_Jobs *, int ));
extern char *SUT_HostName ANSI_ARGS(( int, void *));
#endif
