#include "sut.h"

/*@
   SUT_Pstart - Starts a collection of processes given a list of nodes

   Input Parameters:

   Output Parameters:
.  jobs - Contains information on the created processes.  Must have been
   initialized with SUT_Init.

   Notes:
   This uses a recursive-subdivision approach in starting the processes.
   In all cases, the Manager is started first; this is used to manage the
   data flow to and from the actual program.  The manager is also
   required to respond to a particular startup and exit protocol.

   This only starts the processes.  It returns a structure containing
   the pids, read and write fds, and fdsets for the generated processes.
   The system ensures that the manager process has already started (through
   an exchange of messages), but does no other handling of data.  
   The 'jobs' structure should be given to an appropriate manager routine,
   such as SUT_Passdata.

   In normal use, the executable that this routine calls contains calls
   to SUT_Handshake (to initialize) and to SUT_Pstart (to continue generating
   processes).  In addition, each created job should create a process that 
   manages the actual program to run (or at least forks a process to run
   the user's program).  A manager process may be used to regulate the
   flow of data to and from the user's application, or to add text (for 
   example, that this data is from process 12).

   The executable that is called must interpret the first argument as a
   standard SUT range specification.

   IMPORTANT:
   The first node in the range list MUST be the node that this routine is
   running on.
@*/
int SUT_Pstart( host_nums, count, HostnumToName, HNctx, 
	        ExeNameF, ExeCtx, ArgsF, ArgsC, Environment, jobs )
int       *host_nums, count;
char      *(*HostnumToName)();
void      *HNctx;
char      *(*ExeNameF)();
void      *ExeCtx;
char      **(*ArgsF)();
void      *ArgsC;
char      *Environment;
SUT_Jobs  *jobs;
{
char   **Argv;
int  err, newidx, i;
int  curdepth;/* Current connection number */
int  is,      /* Start of subgroup */
     s,       /* Current start (> is if s, ..., is-1 dead) */
     e;       /* End of subgroup */
SUT_Connection *conn;
SUT_Opfcns     *child_ops = SUT_CreateOpfcns();

/* SUT_Init must be called before calling this routine */

/* If I'm not the first element in host_nums, then there is an 
   error.  NOT YET DETECTED */
jobs->mynodenum = host_nums[0];

/* Current index.  Note that we always skip ourselves (first index) */
is	 = s = (1 + count)/2; 
if (is == 0) is = s = 1;
e	 = count-1;
curdepth = jobs->nconn;
while (s <= e && e > 0) {
    /* Create a new job, complete with pipes for stdin/stdout */
    conn = jobs->conn + curdepth;
    if (err = SUT_Fork_with_pipes( &conn->pid, &conn->readfd, &conn->writefd ))
	return err;

    if (conn->pid == 0) {
	/* Child */
	/* Generate the call to the manager on this remote system */
	
	if (SUT_Get_exec_args( s, ExeNameF, ExeCtx, ArgsF, ArgsC, 
			      Environment, &Argv, 1 ))
	    return SUT_GET_EXE_FAILED;
	/* Set the name of the host */
	Argv[1] = (*HostnumToName)( host_nums[s], HNctx );
	if (!Argv[1]) {
	    fprintf( stderr, "Did not get host name!\n" );
	    }
	/* Set the list of remaining hosts */
	Argv[3] = SUT_Condense( host_nums + s, (e-s+1) );
	
	SUT_execvp( Argv[0], Argv );
	return SUT_EXEC_FAILED;
	}

    /* Save the list of nodes and the index of this child */
    conn->nchild = e - s + 1;
    conn->nlist  = (int *)MALLOC( (e - s + 1) * sizeof(int) );
    for (i=0; i<conn->nchild; i++) {
	conn->nlist[i] = host_nums[s+i];
	}
    jobs->children[jobs->nchildren] = curdepth;

    /* Read the ready message from the created child.
       We could defere this until the first attempts at all of the connections
       have been made.  If a connection fails, the connection data-structure
       contains enough information to run on the remaining part
       of the list */
    switch (err = SUT_Handshake( conn->readfd, conn->writefd, 1 )) {
	case SUT_SUCCESS:
	break;
	
	case SUT_SELECT_FAILED:
	return err;
	
	case SUT_CHILD_UNRESPONSIVE:
	case SUT_CHILD_CONFUSED:
	/* The child is unresponsive.  or confused. 
	   Kill it and go to the next in
	   the list */
	SUT_Kill_child( conn->pid, SIGKILL, "" );
	FREE( conn->nlist );
	s++;
	if (s > e) {
	    /* Entire subtree is down.  Skip to next subtree */
	    e = is - 1;
	    is = s = e / 2;
	    if (is == 0) is = s = 1;
	    }
	close(conn->readfd);
	close(conn->writefd);
	/* Need to skip to the beginning of the while; done with test
	   on err */
	break;
	
	default:
	fprintf( stderr, "Unexpected return from SUT_Handshake\n" );
	break;
	} /* End of switch on select for handshake */
    /* Need code for the case that there are no more hosts in this
       subtree (for example, the LAST member failed) */
    if (!err) {
	/* Now, we skip to the next index to run. */
	conn->nodenum	 = host_nums[s];
	FD_SET(conn->readfd,&jobs->rdset );
	if (conn->readfd > jobs->maxfds) 
	    jobs->maxfds = conn->readfd;
	conn->ops	 = child_ops;
	conn->readdata	 = SUT_ProcessInput;
	curdepth++;
        jobs->nchildren ++;

	/* Implement recursive subdivision */
	e  = is - 1;
	is = s = e / 2;
	if (is == 0) is = s = 1;
	}
    } /* End of while for starting jobs */
jobs->nconn   = curdepth;	
jobs->nactive = curdepth;
return SUT_SUCCESS;
}

/* 
   Start a process with a separate read/write pipe for stdin/stdout, 
   placing the information in the jobs structure.  The entire calling
   vector is in argv;
 */
int SUT_Startprocess( jobs, argv )
SUT_Jobs *jobs;
char     **argv;
{
int err;
int curdepth = jobs->nactive;
SUT_Connection *conn;

conn = jobs->conn + curdepth;

if (err = SUT_Fork_with_pipes( &conn->pid, &conn->readfd, &conn->writefd ))
    return err;

/* Create a new job */
if (conn->pid == 0) {
    /* Child */
    SUT_execvp( argv[0], argv );
    return SUT_EXEC_FAILED;
    } 

conn->readdata	 = SUT_CopyToOut;
FD_SET(conn->readfd,&jobs->rdset );
conn->nodenum	 = SUT_Mynodenum();
if (conn->readfd > jobs->maxfds) 
    jobs->maxfds = conn->readfd;
conn->ops	 = SUT_CreateOpfcns();
conn->handle_eof = SUT_Leaf_exit;
jobs->nconn	 = curdepth + 1;
jobs->nactive	 = curdepth + 1;
jobs->local	 = curdepth;

return SUT_SUCCESS;
}

/* 
   If we are not local, or if we are not sure, call path and perform the
   usual handshake (this allows path to be a slave that KNOWS that it is
   on the correct node) 
 */
int SUT_Notlocal( host_nums, count, HostnumToName, HNctx, path, argv, jobs )
int       *host_nums, count;
char      *(*HostnumToName)();
void      *HNctx;
char      *path;
char      **argv;
SUT_Jobs  *jobs;
{
int            my_num;
int            err;
char           **Argv;
int            cnt, i;
SUT_Connection *conn;

my_num = SUT_Mynodenum();
if (SUT_Debug & SUT_PRINT) 
    fprintf( stderr, "[%d] My nodenumber is %d\n", SUT_Mynode, my_num );
if (host_nums[0] != my_num) {
    if (SUT_Debug & SUT_PRINT)
	fprintf( stderr, "[%d] Switching to first node...\n", SUT_Mynode );
    conn = jobs->conn;
    /* Go to the specified host */
    /* Create a new job */
    if (err = SUT_Fork_with_pipes( &conn->pid, &conn->readfd, 
				   &conn->writefd )) 
	   return err;
    if (conn->pid == 0) {
	/* Child */
	for (cnt=0; argv[cnt]; cnt++) ;
	Argv = (char **)MALLOC( (cnt + 4) * sizeof(char *) );
	if (!Argv) return SUT_NO_MEMORY;
	Argv[0] = SUT_REMSH;
	Argv[1] = (*HostnumToName)( host_nums[0], HNctx );
	if (!Argv[1]) {
	    fprintf( stderr, "Did not get host name!\n" );
	    }
	Argv[2] = path;
	argv++;
	for (i=3; *argv; i++) 
	    Argv[i] = *argv++;
	SUT_execvp( Argv[0], Argv );
	return SUT_EXEC_FAILED;
	}

    /* Read the ready message from the created child */
    switch (err = SUT_Handshake( conn->readfd, conn->writefd, 1 )) {
	case SUT_SUCCESS:
	break;
	
	case SUT_SELECT_FAILED:
	return err;
	
	case SUT_CHILD_UNRESPONSIVE:
	case SUT_CHILD_CONFUSED:
	/* The child is unresponsive.  or confused. 
	   Kill it and go to the next in
	   the list */
	SUT_Kill_child( conn->pid, SIGKILL, "" );
	close(conn->readfd);
	close(conn->writefd);
	break;
	
	default:
	fprintf( stderr, "Unexpected return from SUT_Handshake\n" );
	break;
	} /* End of switch on select for handshake */
    FD_SET(conn->readfd,&jobs->rdset );
    if (conn->readfd > jobs->maxfds) 
	jobs->maxfds = conn->readfd;
    conn->ops	     = SUT_CreateOpfcns();
    conn->readdata   = SUT_ProcessInput;
    jobs->nconn	     = 1;
    jobs->nactive    = 1;
    jobs->nconn	     = 1;
    /* Need to capture and output data */
    return SUT_NOT_LOCAL;
    } /* End of if not host I expected */

return SUT_SUCCESS;
}

/* Add a connection as the parent */
int SUT_AddParent( jobs )
SUT_Jobs *jobs;
{
SUT_Connection *conn = jobs->conn + jobs->nconn;

conn->readfd	 = 0;
conn->writefd	 = 1;
conn->pid	 = -1;     /* Don't know our parent */
conn->nodenum	 = -1;     /* Don't know, but could! */
conn->state	 = SUT_READY;
conn->nchild	 = -1;
conn->nlist	 = 0;
/* Linebuffer already allocated */

/* If our parent is another pexec, we just use ProcessInput */
conn->readdata	 = SUT_ProcessInput;
conn->writedata	 = 0;
conn->handle_eof = 0;   /* Should do a forced exit */
jobs->parent	 = jobs->nconn;
jobs->nconn++;
FD_SET(0,&jobs->rdset);
}

#ifdef FOO2
/* 
   This is a new version that takes an array of host entries
 */
int SUT_Pstart2( entries, count, ExeNameF, ExeCtx, ArgsF, ArgsC, 
		Environment, jobs )
SUT_Node_entry *entries
int       count;
char      *(*ExeNameF)();
void      *ExeCtx;
char      **(*ArgsF)();
void      *ArgsC;
char      *Environment;
SUT_Jobs  *jobs;
{
char   **Argv;
int  err, newidx, i;
int  curdepth;/* Current connection number */
int  is,      /* Start of subgroup */
     s,       /* Current start (> is if s, ..., is-1 dead) */
     e;       /* End of subgroup */
SUT_Connection *conn;
SUT_Opfcns     *child_ops = SUT_CreateOpfcns();

/* SUT_Init must be called before calling this routine */

/* If I'm not the first element in host_nums, then there is an 
   error.  NOT YET DETECTED */
jobs->mynodenum = host_nums[0];

/* Current index.  Note that we always skip ourselves (first index) */
is	 = s = (1 + count)/2; 
if (is == 0) is = s = 1;
e	 = count-1;
curdepth = jobs->nconn;
while (s <= e && e > 0) {
    /* Create a new job, complete with pipes for stdin/stdout */
    conn = jobs->conn + curdepth;
    if (err = SUT_Fork_with_pipes( &conn->pid, &conn->readfd, &conn->writefd ))
	return err;

    if (conn->pid == 0) {
	/* Child */
	/* Generate the call to the manager on this remote system */
	
	if (SUT_Get_exec_args( s, ExeNameF, ExeCtx, ArgsF, ArgsC, 
			      Environment, &Argv, 1 ))
	    return SUT_GET_EXE_FAILED;
	/* Set the name of the host */
	Argv[1] = entries[s].name;
	/* Set the list of remaining hosts */
	/* This is part of the handshake ... */
	/* Argv[3] = SUT_Condense( host_nums + s, (e-s+1) ); */
	
	SUT_execvp( Argv[0], Argv );
	return SUT_EXEC_FAILED;
	}

    /* Save the list of nodes and the index of this child */
    /* Instead, this should either be the range in the Entries array
       or the entries themselves.  Lets use the range... */
    conn->nchild = e - s + 1;
    conn->nlist  = (int *)MALLOC( (e - s + 1) * sizeof(int) );
    for (i=0; i<conn->nchild; i++) {
	conn->nlist[i] = host_nums[s+i];
	}
    jobs->children[jobs->nchildren] = curdepth;

    /* Read the ready message from the created child.
       We could defere this until the first attempts at all of the connections
       have been made.  If a connection fails, the connection data-structure
       contains enough information to run on the remaining part
       of the list */
    switch (err = SUT_Handshake( conn->readfd, conn->writefd, 1 )) {
	case SUT_SUCCESS:
	break;
	
	case SUT_SELECT_FAILED:
	return err;
	
	case SUT_CHILD_UNRESPONSIVE:
	case SUT_CHILD_CONFUSED:
	/* The child is unresponsive.  or confused. 
	   Kill it and go to the next in
	   the list */
	SUT_Kill_child( conn->pid, SIGKILL, "" );
	FREE( conn->nlist );
	s++;
	if (s > e) {
	    /* Entire subtree is down.  Skip to next subtree */
	    e = is - 1;
	    is = s = e / 2;
	    if (is == 0) is = s = 1;
	    }
	close(conn->readfd);
	close(conn->writefd);
	/* Need to skip to the beginning of the while; done with test
	   on err */
	break;
	
	default:
	fprintf( stderr, "Unexpected return from SUT_Handshake\n" );
	break;
	} /* End of switch on select for handshake */
    /* Need code for the case that there are no more hosts in this
       subtree (for example, the LAST member failed) */
    if (!err) {
	/* Now, we skip to the next index to run. */
	conn->nodenum	 = host_nums[s];
	FD_SET(conn->readfd,&jobs->rdset );
	if (conn->readfd > jobs->maxfds) 
	    jobs->maxfds = conn->readfd;
	conn->ops	 = child_ops;
	conn->readdata	 = SUT_ProcessInput;
	curdepth++;
        jobs->nchildren ++;

	/* Implement recursive subdivision */
	e  = is - 1;
	is = s = e / 2;
	if (is == 0) is = s = 1;
	}
    } /* End of while for starting jobs */
jobs->nconn   = curdepth;	
jobs->nactive = curdepth;
return SUT_SUCCESS;
}
#endif
